#include <alloca.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include <DHT.h>
#include <RFTransmitter.h>
#include "Voltmeter.h"
#include "SensorData.h"
#include "Sleep.h"

#define NODE_ID             2

// Hardware Revision 1
/*
#define DOUBLER_POWER_PIN   6
#define DOUBLER_PWM_PIN     7
*/

// Hardware Revision 2
#define DOUBLER_POWER_PIN   7
#define DOUBLER_PWM_PIN     8
#define DOUBLER_THRESHOLD 3400

RFTransmitter transmitter(1, NODE_ID);

OneWire oneWire(10);
DallasTemperature dallas(&oneWire);
DHT dht(2, DHT22);


static void transmitterOn() {
  pinMode(0, OUTPUT);
  digitalWrite(0, HIGH);
}

static void transmitterOff() {
  pinMode(0, INPUT);
  digitalWrite(0, LOW);
}

static void sensorsOff() {
  pinMode(9, INPUT);
  digitalWrite(9, LOW);
  doublerOff();
}

static void doublerOn() {
  pinMode(DOUBLER_POWER_PIN, OUTPUT);
  digitalWrite(DOUBLER_POWER_PIN, HIGH);
  pinMode(DOUBLER_PWM_PIN, OUTPUT);
  analogWrite(DOUBLER_PWM_PIN, 127);
}

static void doublerOff() {
  pinMode(DOUBLER_PWM_PIN, INPUT);
  digitalWrite(DOUBLER_PWM_PIN, LOW);
  pinMode(DOUBLER_POWER_PIN, INPUT);
  digitalWrite(DOUBLER_POWER_PIN, LOW);
}

static void sensorsOn(bool useDoubler = false) {
  if (!useDoubler) {
    doublerOff();

    pinMode(9, OUTPUT);
    digitalWrite(9, HIGH);
  } else {
    pinMode(9, INPUT);
    digitalWrite(9, LOW);
    
    doublerOn();
  }
  
  delay(50);

  dht.begin();
  dallas.begin();
}

#define wdt_reset() __asm__ __volatile__ ("wdr")

static void wdt_disable() {
  // Disable Watchdog
  noInterrupts();
  
  wdt_reset();
  /* Clear WDRF in MCUSR */
  MCUSR = 0;
  // Do not optimize!
  WDTCSR |= (1 << WDCE) | (1 << WDE);
  WDTCSR = 0;

  interrupts();
}

static void wdt_enable() {
  // Normal wdt_enable() does not work properly
  noInterrupts();
  
  // Clear the reset flag
  MCUSR &= ~(1 << WDRF);
  
  /*
   * In order to change WDE or the prescaler, we need to
   * set WDCE (This will allow updates for 4 clock cycles).
   */
  WDTCSR |= (1 << WDCE) | (1 << WDE);

  // set new watchdog timeout prescaler value
  // 8.0 seconds
  WDTCSR = 1 << WDP0 | 1 << WDP3;
  
  WDTCSR |= _BV(WDE);

  interrupts();
}

static void sendSensorData() {
  unsigned int unloadedVcc;
  byte chargingCounter = 0;
  byte numberOfDevices, dataSize;
  SensorData *data;

  unloadedVcc = readVcc();

  transmitterOn();
  sensorsOn(unloadedVcc < DOUBLER_THRESHOLD);
  delay(500);
  
  numberOfDevices = dallas.getDeviceCount();
  dataSize = sizeof(SensorData) + DALLAS_SIZE * numberOfDevices;
  data = (SensorData *) alloca(dataSize);

  for (;;) {
    data->unloadedVcc = unloadedVcc;
    data->vcc = readVcc();

    data->dht.humidity = dht.readHumidity();
    data->dht.temperature = dht.readTemperature();
    
    data->dallasCount = numberOfDevices;
    dallas.requestTemperatures();

    wdt_reset();
  
    for(byte i = 0; i < numberOfDevices; ++i) {
      if(dallas.getAddress(data->dallas[i].address, i)) {
        dallas.setResolution(data->dallas[i].address, 12);
        data->dallas[i].temperature = dallas.getTempC(data->dallas[i].address);
      } else {
        memset(data->dallas[i].address, 0, sizeof(DeviceAddress));
        data->dallas[i].temperature = NAN;
      }
    }

    wdt_reset();
  
    sensorsOff();

    transmitter.send((byte *)data, dataSize);

    wdt_disable();

    sleep(8);

    wdt_enable();
    
    transmitter.resend((byte *)data, dataSize);

    transmitterOff();

    wdt_disable();
    
    if (data->vcc > 4200 || chargingCounter >= 5)
      sleep(16);
    else if (data->vcc > 3000)
      sleep(150);
    else if (data->vcc > 2000)
      sleep(300);
    else {
      do {
        sleep(600);
      } while (readVcc() <= 2000);
    }

    wdt_enable();
    
    unloadedVcc = readVcc();
    // If the supply voltage is increasing, then
    // batteries or capacitors are charging
    if (unloadedVcc > data->unloadedVcc)
      if (chargingCounter < 10)
        ++chargingCounter;
    else
      chargingCounter = 0;

    sensorsOn(unloadedVcc < DOUBLER_THRESHOLD);
    transmitterOn();

    delay(250);

    transmitter.resend((byte *)data, dataSize);
    wdt_reset();
  }
}

void setup() {
  // Very important to prevent crash
  wdt_disable();
  
  randomSeed(analogRead(0));

  doublerOff();
}

void loop() {
  sendSensorData();
}
